@@ -37,20 +37,26 @@ https://github.com/microsoft/webauthn
 #
 # pylint: disable=invalid-name, super-init-not-called, too-few-public-methods
 
-from __future__ import absolute_import, unicode_literals
+from __future__ import annotations
 
 from enum import IntEnum, unique
 from ctypes.wintypes import BOOL, DWORD, LONG, LPCWSTR, HWND
 from threading import Thread
+from typing import Mapping
 
 import ctypes
+from ctypes import WinDLL  # type: ignore
+from ctypes import LibraryLoader
+
+
+windll = LibraryLoader(WinDLL)
 
 
 PBYTE = ctypes.POINTER(ctypes.c_ubyte)  # Different from wintypes.PBYTE, which is signed
 PCWSTR = ctypes.c_wchar_p
 
 
-class BytesProperty(object):
+class BytesProperty:
     """Property for structs storing byte arrays as DWORD + PBYTE.
 
     Allows for easy reading/writing to struct fields using Python bytes objects.
@@ -86,10 +92,10 @@ class GUID(ctypes.Structure):
             self.Data2,
             self.Data3,
             self.Data4[0] * 256 + self.Data4[1],
-            self.Data4[2] * (256 ** 5)
-            + self.Data4[3] * (256 ** 4)
-            + self.Data4[4] * (256 ** 3)
-            + self.Data4[5] * (256 ** 2)
+            self.Data4[2] * (256**5)
+            + self.Data4[3] * (256**4)
+            + self.Data4[4] * (256**3)
+            + self.Data4[5] * (256**2)
             + self.Data4[6] * 256
             + self.Data4[7],
         )
@@ -347,8 +353,13 @@ class WebAuthNGetAssertionOptions(ctypes.Structure):
         ("pbU2fAppId", BOOL),
         ("pCancellationId", ctypes.POINTER(GUID)),
         ("pAllowCredentialList", ctypes.POINTER(WebAuthNCredentialList)),
+        ("dwCredLargeBlobOperation", DWORD),
+        ("cbCredLargeBlob", DWORD),
+        ("pbCredLargeBlob", PBYTE),
     ]
 
+    cred_large_blob = BytesProperty("CredLargeBlob")
+
     def __init__(
         self,
         timeout,
@@ -356,6 +367,8 @@ class WebAuthNGetAssertionOptions(ctypes.Structure):
         user_verification_requirement,
         credentials,
         cancellationId,
+        cred_large_blob_operation,
+        cred_large_blob,
     ):
         self.dwVersion = get_version(self.__class__.__name__)
         self.dwTimeoutMilliseconds = timeout
@@ -371,6 +384,10 @@ class WebAuthNGetAssertionOptions(ctypes.Structure):
         else:
             self.CredentialList = WebAuthNCredentials(credentials)
 
+        if self.dwVersion >= 5:
+            self.cred_large_blob_operation = cred_large_blob_operation
+            self.cred_large_blob = cred_large_blob
+
 
 class WebAuthNAssertion(ctypes.Structure):
     """Maps to WEBAUTHN_ASSERTION Struct.
@@ -387,11 +404,16 @@ class WebAuthNAssertion(ctypes.Structure):
         ("Credential", WebAuthNCredential),
         ("cbUserId", DWORD),
         ("pbUserId", PBYTE),
+        ("Extensions", WebAuthNExtensions),
+        ("cbCredLargeBlob", DWORD),
+        ("pbCredLargeBlob", PBYTE),
+        ("dwCredLargeBlobStatus", DWORD),
     ]
 
     auth_data = BytesProperty("AuthenticatorData")
     signature = BytesProperty("Signature")
     user_id = BytesProperty("UserId")
+    cred_large_blob = BytesProperty("CredLargeBlob")
 
     def __del__(self):
         WEBAUTHN.WebAuthNFreeAssertion(ctypes.byref(self))
@@ -426,6 +448,9 @@ class WebAuthNMakeCredentialOptions(ctypes.Structure):
         ("dwFlags", DWORD),
         ("pCancellationId", ctypes.POINTER(GUID)),
         ("pExcludeCredentialList", ctypes.POINTER(WebAuthNCredentialList)),
+        ("dwEnterpriseAttestation", DWORD),
+        ("dwLargeBlobSupport", DWORD),
+        ("bPreferResidentKey", BOOL),
     ]
 
     def __init__(
@@ -437,6 +462,9 @@ class WebAuthNMakeCredentialOptions(ctypes.Structure):
         attestation_convoyence,
         credentials,
         cancellationId,
+        enterprise_attestation,
+        large_blob_support,
+        prefer_resident_key,
     ):
         self.dwVersion = get_version(self.__class__.__name__)
         self.dwTimeoutMilliseconds = timeout
@@ -455,6 +483,11 @@ class WebAuthNMakeCredentialOptions(ctypes.Structure):
         else:
             self.CredentialList = WebAuthNCredentials(credentials)
 
+        if self.dwVersion >= 4:
+            self.dwEnterpriseAttestation = enterprise_attestation
+            self.dwLargeBlobSupport = large_blob_support
+            self.bPreferResidentKey = prefer_resident_key
+
 
 class WebAuthNCredentialAttestation(ctypes.Structure):
     """Maps to WEBAUTHN_CREDENTIAL_ATTESTATION Struct.
@@ -477,6 +510,9 @@ class WebAuthNCredentialAttestation(ctypes.Structure):
         ("pbCredentialId", PBYTE),
         ("Extensions", WebAuthNExtensions),
         ("dwUsedTransport", DWORD),
+        ("bEpAtt", BOOL),
+        ("bLargeBlobSupported", BOOL),
+        ("bResidentKey", BOOL),
     ]
 
     auth_data = BytesProperty("AuthenticatorData")
@@ -488,8 +524,14 @@ class WebAuthNCredentialAttestation(ctypes.Structure):
         WEBAUTHN.WebAuthNFreeCredentialAttestation(ctypes.byref(self))
 
 
+class _FromString(object):
+    @classmethod
+    def from_string(cls, value):
+        return getattr(cls, value.upper().replace("-", "_"))
+
+
 @unique
-class WebAuthNUserVerificationRequirement(IntEnum):
+class WebAuthNUserVerificationRequirement(_FromString, IntEnum):
     """Maps to WEBAUTHN_USER_VERIFICATION_REQUIREMENT_*.
 
     https://github.com/microsoft/webauthn/blob/master/webauthn.h#L335
@@ -500,13 +542,9 @@ class WebAuthNUserVerificationRequirement(IntEnum):
     PREFERRED = 2
     DISCOURAGED = 3
 
-    @classmethod
-    def from_string(cls, value):
-        return getattr(cls, value.upper().replace("-", "_"))
-
 
 @unique
-class WebAuthNAttestationConvoyancePreference(IntEnum):
+class WebAuthNAttestationConvoyancePreference(_FromString, IntEnum):
     """Maps to WEBAUTHN_ATTESTATION_CONVEYANCE_PREFERENCE_*.
 
     https://github.com/microsoft/webauthn/blob/master/webauthn.h#L340
@@ -517,13 +555,9 @@ class WebAuthNAttestationConvoyancePreference(IntEnum):
     INDIRECT = 2
     DIRECT = 3
 
-    @classmethod
-    def from_string(cls, value):
-        return getattr(cls, value.upper().replace("-", "_"))
-
 
 @unique
-class WebAuthNAuthenticatorAttachment(IntEnum):
+class WebAuthNAuthenticatorAttachment(_FromString, IntEnum):
     """Maps to WEBAUTHN_AUTHENTICATOR_ATTACHMENT_*.
 
     https://github.com/microsoft/webauthn/blob/master/webauthn.h#L330
@@ -534,13 +568,9 @@ class WebAuthNAuthenticatorAttachment(IntEnum):
     CROSS_PLATFORM = 2
     CROSS_PLATFORM_U2F_V2 = 3
 
-    @classmethod
-    def from_string(cls, value):
-        return getattr(cls, value.upper().replace("-", "_"))
-
 
 @unique
-class WebAuthNCTAPTransport(IntEnum):
+class WebAuthNCTAPTransport(_FromString, IntEnum):
     """Maps to WEBAUTHN_CTAP_TRANSPORT_*.
 
     https://github.com/microsoft/webauthn/blob/master/webauthn.h#L225
@@ -554,12 +584,46 @@ class WebAuthNCTAPTransport(IntEnum):
     INTERNAL = 0x00000010
     FLAGS_MASK = 0x0000001F
 
-    @classmethod
-    def from_string(cls, value):
-        return getattr(cls, value.upper().replace("-", "_"))
+
+@unique
+class WebAuthNEnterpriseAttestation(_FromString, IntEnum):
+    """Maps to WEBAUTHN_ENTERPRISE_ATTESTATION_*.
+
+    https://github.com/microsoft/webauthn/blob/master/webauthn.h#L401
+    """
+
+    NONE = 0
+    VENDOR_FACILITATED = 1
+    PLATFORM_MANAGED = 2
 
 
-WEBAUTHN = ctypes.windll.webauthn
+@unique
+class WebAuthNLargeBlobSupport(_FromString, IntEnum):
+    """Maps to WEBAUTHN_LARGE_BLOB_SUPPORT_*.
+
+    https://github.com/microsoft/webauthn/blob/master/webauthn.h#L405
+    """
+
+    NONE = 0
+    REQUIRED = 1
+    PREFERRED = 2
+
+
+@unique
+class WebAuthNLargeBlobOperation(_FromString, IntEnum):
+    """Maps to WEBAUTHN_LARGE_BLOB_OPERATION_*.
+
+    https://github.com/microsoft/webauthn/blob/master/webauthn.h#L478
+    """
+
+    NONE = 0
+    GET = 1
+    SET = 2
+    DELETE = 3
+
+
+HRESULT = ctypes.HRESULT  # type: ignore
+WEBAUTHN = windll.webauthn  # type: ignore
 WEBAUTHN_API_VERSION = WEBAUTHN.WebAuthNGetApiVersionNumber()
 # The following is derived from
 # https://github.com/microsoft/webauthn/blob/master/webauthn.h#L37
@@ -567,7 +631,7 @@ WEBAUTHN_API_VERSION = WEBAUTHN.WebAuthNGetApiVersionNumber()
 WEBAUTHN.WebAuthNIsUserVerifyingPlatformAuthenticatorAvailable.argtypes = [
     ctypes.POINTER(ctypes.c_bool)
 ]
-WEBAUTHN.WebAuthNIsUserVerifyingPlatformAuthenticatorAvailable.restype = ctypes.HRESULT
+WEBAUTHN.WebAuthNIsUserVerifyingPlatformAuthenticatorAvailable.restype = HRESULT
 
 WEBAUTHN.WebAuthNAuthenticatorMakeCredential.argtypes = [
     HWND,
@@ -578,7 +642,7 @@ WEBAUTHN.WebAuthNAuthenticatorMakeCredential.argtypes = [
     ctypes.POINTER(WebAuthNMakeCredentialOptions),
     ctypes.POINTER(ctypes.POINTER(WebAuthNCredentialAttestation)),
 ]
-WEBAUTHN.WebAuthNAuthenticatorMakeCredential.restype = ctypes.HRESULT
+WEBAUTHN.WebAuthNAuthenticatorMakeCredential.restype = HRESULT
 
 WEBAUTHN.WebAuthNAuthenticatorGetAssertion.argtypes = [
     HWND,
@@ -587,7 +651,7 @@ WEBAUTHN.WebAuthNAuthenticatorGetAssertion.argtypes = [
     ctypes.POINTER(WebAuthNGetAssertionOptions),
     ctypes.POINTER(ctypes.POINTER(WebAuthNAssertion)),
 ]
-WEBAUTHN.WebAuthNAuthenticatorGetAssertion.restype = ctypes.HRESULT
+WEBAUTHN.WebAuthNAuthenticatorGetAssertion.restype = HRESULT
 
 WEBAUTHN.WebAuthNFreeCredentialAttestation.argtypes = [
     ctypes.POINTER(WebAuthNCredentialAttestation)
@@ -595,16 +659,16 @@ WEBAUTHN.WebAuthNFreeCredentialAttestation.argtypes = [
 WEBAUTHN.WebAuthNFreeAssertion.argtypes = [ctypes.POINTER(WebAuthNAssertion)]
 
 WEBAUTHN.WebAuthNGetCancellationId.argtypes = [ctypes.POINTER(GUID)]
-WEBAUTHN.WebAuthNGetCancellationId.restype = ctypes.HRESULT
+WEBAUTHN.WebAuthNGetCancellationId.restype = HRESULT
 
 WEBAUTHN.WebAuthNCancelCurrentOperation.argtypes = [ctypes.POINTER(GUID)]
-WEBAUTHN.WebAuthNCancelCurrentOperation.restype = ctypes.HRESULT
+WEBAUTHN.WebAuthNCancelCurrentOperation.restype = HRESULT
 
-WEBAUTHN.WebAuthNGetErrorName.argtypes = [ctypes.HRESULT]
+WEBAUTHN.WebAuthNGetErrorName.argtypes = [HRESULT]
 WEBAUTHN.WebAuthNGetErrorName.restype = PCWSTR
 
 
-WEBAUTHN_STRUCT_VERSIONS = {
+WEBAUTHN_STRUCT_VERSIONS: Mapping[int, Mapping[str, int]] = {
     1: {
         "WebAuthNRpEntityInformation": 1,
         "WebAuthNUserEntityInformation": 1,
@@ -614,30 +678,39 @@ WEBAUTHN_STRUCT_VERSIONS = {
         "WebAuthNCredentialEx": 1,
         "WebAuthNMakeCredentialOptions": 3,
         "WebAuthNGetAssertionOptions": 4,
-        "WEBAUTHN_COMMON_ATTESTATION": 1,
+        "WebAuthNCommonAttestation": 1,
         "WebAuthNCredentialAttestation": 3,
         "WebAuthNAssertion": 1,
     },
     2: {},
+    3: {
+        "WebAuthNMakeCredentialOptions": 4,
+        "WebAuthNGetAssertionOptions": 5,
+        "WebAuthNCredentialAttestation": 4,
+        "WebAuthNAssertion": 2,
+    },
 }
 
 
-def get_version(class_name):
+def get_version(class_name: str) -> int:
     """Get version of struct.
 
     :param str class_name: Struct class name.
     :returns: Version of Struct to use.
     :rtype: int
     """
-    if class_name in WEBAUTHN_STRUCT_VERSIONS[WEBAUTHN_API_VERSION]:
-        return WEBAUTHN_STRUCT_VERSIONS[WEBAUTHN_API_VERSION][class_name]
-
-    return WEBAUTHN_STRUCT_VERSIONS[1][class_name]
+    for api_version in range(WEBAUTHN_API_VERSION, 0, -1):
+        if (
+            api_version in WEBAUTHN_STRUCT_VERSIONS
+            and class_name in WEBAUTHN_STRUCT_VERSIONS[api_version]
+        ):
+            return WEBAUTHN_STRUCT_VERSIONS[api_version][class_name]
+    raise ValueError("Unknown class name")
 
 
 class CancelThread(Thread):
     def __init__(self, event):
-        super(CancelThread, self).__init__()
+        super().__init__()
         self.daemon = True
         self._completed = False
         self.event = event
@@ -655,7 +728,7 @@ class CancelThread(Thread):
         self.join()
 
 
-class WinAPI(object):
+class WinAPI:
     """Implementation of Microsoft's WebAuthN APIs.
 
     :param ctypes.HWND handle: Window handle to use for API calls.
@@ -664,7 +737,7 @@ class WinAPI(object):
     version = WEBAUTHN_API_VERSION
 
     def __init__(self, handle=None):
-        self.handle = handle or ctypes.windll.user32.GetForegroundWindow()
+        self.handle = handle or windll.user32.GetForegroundWindow()
 
     def get_error_name(self, winerror):
         """Returns an error name given an error HRESULT value.
@@ -738,6 +811,10 @@ class WinAPI(object):
                     attestation,
                     exclude_credentials or [],
                     ctypes.pointer(t.guid) if event else None,
+                    # Not supported:
+                    WebAuthNEnterpriseAttestation.NONE,
+                    WebAuthNLargeBlobSupport.NONE,
+                    False,
                 )
             ),
             ctypes.byref(attestation_pointer),
@@ -790,6 +867,9 @@ class WinAPI(object):
                     user_verification,
                     allow_credentials or [],
                     ctypes.pointer(t.guid) if event else None,
+                    # Not supported:
+                    WebAuthNLargeBlobOperation.NONE,
+                    b"",
                 )
             ),
             ctypes.byref(assertion_pointer),
